(ns wanijo.framework.form
  (:require [clojure.spec.alpha :as spec]
            [hiccup.form :as hform]
            [hiccup.core :as hcore]
            [wanijo.framework.view :as view]
            [ring.util.anti-forgery :refer [anti-forgery-field]]
            [wanijo.framework.common :refer [in?]]))

(spec/def ::label string?)
(spec/def ::required boolean?)
(spec/def ::spec keyword?)
(spec/def ::options
  (spec/or :empty empty?
           :options (spec/coll-of (spec/tuple string? string?))))
(spec/def ::widget
  #(in? [:input :select :checkbox :textarea :mselect] %))
(spec/def ::field
  (spec/keys :req-un [::label
                      ::required
                      ::spec]
             :opt-un [::options
                      ::widget]))
(spec/def ::fields
  (spec/map-of keyword? ::field))
(spec/def ::form
  (spec/keys :req-un [::fields]))

(defn spec-to-errmsg [label spec-key field-value]
  (view/flash-error
   (map
    (fn [prob]
      [:p
       "Field "
       [:span.flash__field label]
       " must comply to "
       [:span.flash__pred (:pred prob)]])
    (:clojure.spec.alpha/problems
     (spec/explain-data spec-key field-value)))))

(defn field-valid?[value spec-key req]
  (or (empty? (:form-params req))
      (spec/valid? spec-key value)))

(defn field
  ([form-def field-id req]
   (field form-def field-id req
          (get-in req [:params field-id])))
  ([form-def field-id req field-value]
   {:pre [(spec/valid? ::form form-def)]}
   (let [field-def (get-in form-def [:fields field-id])
         {:keys [label required] spec-key :spec} field-def]
     (list (when-not (field-valid? field-value spec-key req)
             (spec-to-errmsg label spec-key field-value))
           (hform/label field-id label)
           (hform/text-field {:required (when required "required")}
                             field-id
                             field-value)))))

(defn readonly
  [form-def field-id field-value]
  (list (hform/label field-id (get-in form-def [:fields field-id :label]))
        (hform/text-field {:readonly "readonly"}
                          field-id
                          field-value)))

(defn valid? [form-def req]
  {:pre [(spec/valid? ::form form-def)]}
  (reduce-kv
   (fn [result field value]
     (if-let [field-spec (get-in form-def [:fields field :spec])]
       (if (spec/valid? field-spec value)
         true
         (reduced false))
       result))
   true
   (:params req)))

(defn render-input [id {:keys [label required spec]} value req-value validate?]
  (list
   (when (and validate? (not (spec/valid? spec req-value)))
     (spec-to-errmsg label spec req-value))
   (hform/label id label)
   (hform/text-field {:required (when required "required")}
                     id
                     (if validate? req-value value))))

(defn render-checkbox [id {:keys [label spec]} value req-value validate?]
  (list
   (when (and validate? (not (spec/valid? spec req-value)))
     (spec-to-errmsg label spec value))
   (hform/label id label)
   (hform/check-box id
                    (and (some? value) (not= 0 value))
                    (if validate? req-value value))))

(defn render-textarea [id {:keys [label spec]} value req-value validate?]
  (list
   (when (and validate? (not (spec/valid? spec req-value)))
     (spec-to-errmsg label spec value))
   (hform/label id label)
   (hform/text-area id
                    (if validate? req-value value))))

(defn render-select [id {:keys [label spec options]} value req-value validate?]
  (list
   (when (and validate? (not (spec/valid? spec req-value)))
     (spec-to-errmsg label spec value))
   (hform/label id label)
   (hform/drop-down id
                    options
                    (if validate? req-value value))))

(defn render-multiselect [id
                          {:keys [label spec options]}
                          value
                          req-value
                          validate?]
  (list
   (when (and validate? (not (spec/valid? spec req-value)))
     (spec-to-errmsg label spec value))
   (hform/label id label)
   (hcore/html
    [:select {:multiple "multiple"
              :size 5
              :name (name id)
              :id (name id)}
     (for [option options]
       [:option {:value (second option)
                 :selected (in? (if validate? req-value value)
                                (second option))}
        (first option)])])))

(def render-mapping
  {:input render-input
   :checkbox render-checkbox
   :textarea render-textarea
   :select render-select
   :mselect render-multiselect})

(defn render-widgets [form-def values req]
  {:pre [(spec/assert ::form form-def)]}
  (let [form-hash (str (hash [form-def values]))
        submitted-hash (get-in req [:params :__form-hash])
        validate? (= form-hash submitted-hash)]
    (conj (map
           (fn [[field-id field-def]]
             (let [{:keys [options widget]} field-def
                   value (field-id values)
                   req-value (get-in req [:params field-id])
                   widget (cond (some? widget) widget
                                (some? options) :select
                                :else :input)
                   widget-renderer (widget render-mapping)]
               (widget-renderer field-id
                                field-def
                                value
                                req-value
                                validate?)))
           (:fields form-def))
          (hform/hidden-field "__form-hash" form-hash)
          (anti-forgery-field))))
