(ns wanijo.schema.domain
  (:require [clojure.spec.alpha :as spec]
            [wanijo.framework.neo4j :as neo4j]))

(spec/def ::name
  (spec/and string? not-empty))

(spec/def ::created-at
  (spec/and string? ::neo4j/date-str))

(spec/def ::assigned-to
  (spec/or :public empty?
           :assigned (spec/coll-of string?)
           :single string?))

(neo4j/defquery
  all-created-by
  "MATCH (s:schema)-[:created_by]->(u:user)
   WHERE u.uuid = {uuid}
   RETURN s
   ORDER BY s.name")

(defn all-created-by! [user-uuid]
  (->>
   (neo4j/exec-query!
    all-created-by
    {:uuid user-uuid})
   (map :s)))

(neo4j/defquery
  all
  "MATCH (s:schema)
   RETURN s
   ORDER BY s.name")

(defn all! []
  (map :s
       (neo4j/exec-query! all {})))

(neo4j/defquery
  create-new
  "MATCH (u:user)
   WHERE u.uuid = {u_uuid}
   CREATE (s:schema)-[:created_by]->(u)
   SET s.name = {name}
   SET s.uuid = {s_uuid}
   SET s.created_at = {created_at}
   RETURN s")

(defn create-new! [schema-name user-uuid]
  (->>
   (neo4j/exec-query!
    create-new
    {:u_uuid user-uuid
     :name schema-name
     :s_uuid (neo4j/uuid)
     :created_at (neo4j/now-str)})
   first
   :uuid))

(neo4j/defquery
  find-by-uuid
  "MATCH (s:schema)
   WHERE s.uuid = {uuid}
   RETURN s")

(defn find-by-uuid! [uuid]
  (->>
   (neo4j/exec-query!
    find-by-uuid
    {:uuid uuid})
   first
   :s))

(neo4j/defquery
  schema-creator
  "MATCH (s:schema)-->(u:user)
   WHERE s.uuid = {uuid}
   RETURN u.uuid as uuid")

(defn find-schema-creator! [uuid]
  (->>
   (neo4j/exec-query!
    schema-creator
    {:uuid uuid})
   first
   :uuid))

(defn can-user-modify? [schema-uuid user-uuid]
  (let [creator (find-schema-creator! schema-uuid)]
    (= creator user-uuid)))

(neo4j/defquery
  delete
  "MATCH (s:schema)
   WHERE s.uuid = {uuid}
   OPTIONAL MATCH (s)--(a:attribute)
   DELETE a
   WITH s
   MATCH (s)-[c]-()
   DELETE c, s")

(defn delete! [uuid]
  (neo4j/exec-query!
   delete
   {:uuid uuid}))

(neo4j/defquery
  edit
  "MATCH (s:schema)
   WHERE s.uuid = {uuid}
     SET s.name = {name}")

(defn edit! [schema]
  (neo4j/exec-query!
   edit
   schema))

(neo4j/defquery
  assigned-users
  "MATCH (s:schema)-[:assigned_to]-(u:user)
   WHERE s.uuid = {uuid}
   RETURN u
   ORDER BY u.ident")

(defn assigned-users! [uuid]
  (map :u
       (neo4j/exec-query!
        assigned-users
        {:uuid uuid})))

(defn find-with-assigned-users! [uuid]
  (assoc (find-by-uuid! uuid)
          :assigned (map :uuid (assigned-users! uuid))))

(neo4j/defquery
  remove-assignments
  "MATCH (s:schema)-[c:assigned_to]-(:user)
   WHERE s.uuid = {uuid}
   DELETE c")

(neo4j/defquery
  create-assignment
  "MATCH (s:schema), (u:user)
   WHERE s.uuid = {uuid}
     AND u.uuid IN {users}
   CREATE (s)-[:assigned_to]->(u)")

(defn assign-users! [uuid users]
  (neo4j/exec-query!
   remove-assignments
   {:uuid uuid})
  (neo4j/exec-query!
   create-assignment
   {:uuid uuid
    :users users}))
